from __future__ import annotations

import copy
import enum
import importlib.metadata
import logging
import os
import sys
import time
from contextlib import asynccontextmanager
from datetime import datetime, timezone
from functools import partial
from os import PathLike
from ssl import SSLContext
from typing import AsyncContextManager, Dict, Optional, Union

import anyio
import attr
from anyio import CancelScope, EndOfStream, Lock, WouldBlock, sleep
from anyio.abc import ByteStream, TaskGroup
from anyio.lowlevel import checkpoint

from serena.channel import Channel
from serena.enums import ClassID, ReplyCode
from serena.exc import (
    AMQPError,
    AMQPStateError,
    InvalidPayloadTypeError,
    InvalidProtocolError,
    UnexpectedCloseError,
)
from serena.frame import Frame, FrameType
from serena.frameparser import NEED_DATA, FrameParser
from serena.payloads.header import BasicHeader
from serena.payloads.method import (
    BasicDeliverPayload,
    BasicGetEmptyPayload,
    BasicGetOkPayload,
    BasicQOSOkPayload,
    BasicQOSPayload,
    ChannelCloseOkPayload,
    ChannelClosePayload,
    ChannelOpenPayload,
    ConfirmSelectOkPayload,
    ConfirmSelectPayload,
    ConnectionCloseOkPayload,
    ConnectionClosePayload,
    ConnectionOpenOkPayload,
    ConnectionOpenPayload,
    ConnectionStartOkPayload,
    ConnectionStartPayload,
    ConnectionTuneOkPayload,
    ConnectionTunePayload,
    MethodFrame,
    MethodPayload,
    method_payload_name,
)
from serena.utils.bitset import BitSet

logger = logging.getLogger(__name__)


class AMQPState(enum.IntEnum):
    """
    State enumeration for the connection object.
    """

    #: The initial state. The connection has opened and we are waiting for the Start payload.
    INITIAL = 0

    #: We have received the Start payload, and we are waiting for the Tune payload.
    RECEIVED_START = 1

    #: We have received the Tune payload, and we are waiting for the Open-Ok payload.
    RECEIVED_TUNE = 2


@attr.s(slots=True, frozen=False)
class HeartbeatStatistics:
    #: The number of heartbeats so far.
    heartbeat_count: int = attr.ib(default=0)

    #: The previous heartbeat time, in monotonic nanoseconds.
    prev_heartbeat_mn: int = attr.ib(default=None)

    #: The current heartbeat time, in monotonic nanoseconds.
    cur_heartbeat_mn: int = attr.ib(default=None)

    #: The previous heartbeat time, in wall clock time.
    prev_heartbeat_time: datetime = attr.ib(default=None)

    #: The current heartbeat time, in wall clock time.
    cur_heartbeat_time: datetime = attr.ib(default=None)

    @property
    def interval(self) -> Optional[int]:
        """
        Returns the interval between two heartbeats in nanoseconds.
        """

        if self.prev_heartbeat_mn is None:
            return None

        return self.cur_heartbeat_mn - self.prev_heartbeat_mn

    def update(self):
        self.heartbeat_count += 1

        self.prev_heartbeat_mn = self.cur_heartbeat_mn
        self.prev_heartbeat_time = self.cur_heartbeat_time

        self.cur_heartbeat_mn = time.monotonic_ns()
        self.cur_heartbeat_time = datetime.now(timezone.utc).astimezone()


# noinspection PyProtectedMember
class AMQPConnection(object):
    """
    A single AMQP connection.
    """

    def __init__(
        self,
        stream: ByteStream,
        *,
        heartbeat_interval: int = 60,
        channel_buffer_size: int = 48,  # reasonable default
        desired_frame_size: int = 131072,
    ):
        """
        :param stream: The :class:`.ByteStream` to use.
        :param heartbeat_interval: The heartbeat interval to negotiate with, in seconds.
        :param channel_buffer_size: The buffer size for channel messages.
        :param desired_frame_size: The maximum body message frame size desired.
        """

        self._sock = stream
        self._parser = FrameParser()
        self._heartbeat_interval = heartbeat_interval
        self._channel_buffer_size = channel_buffer_size
        self._cancel_scope: CancelScope = None  # type: ignore

        self._closed = False
        self._server_requested_close = False
        self._close_info: Optional[ConnectionClosePayload] = None

        self._is_rabbitmq = False

        # negotiated data
        self._actual_heartbeat_interval = 0
        self._max_frame_size = desired_frame_size
        self._server_capabilites = {}

        # list of 64-bit ints (as not to overflow long values and cause a bigint)
        # that is used to assign the next channel ID
        self._channels: BitSet = BitSet(0)

        # mapping of channel id -> Channel
        self._channel_channels: Dict[int, Channel] = {}

        self._write_lock = Lock()

        # statistics
        self._heartbeat_stats = HeartbeatStatistics()

    @staticmethod
    def get_client_properties():
        version = importlib.metadata.version("serena")

        return {
            "product": "Serena AMQP client".encode("utf-8"),
            "platform": f"Python {sys.version}".encode("utf-8"),
            "version": version.encode("utf-8"),
            "capabilities": {
                "publisher_confirms": True,
                "basic.nack": True,
                "authentication_failure_close": True,
                "per_consumer_qos": True,
            },
        }

    @property
    def open(self) -> bool:
        """
        Checks if this connection is actually open.
        """

        return not self._closed and not self._server_requested_close

    def has_capability(self, name: str) -> bool:
        """
        Checks if the server exposes a capability.
        """

        return self._server_capabilites.get(name, False)

    def heartbeat_statistics(self) -> HeartbeatStatistics:
        """
        Returns a copy of the heartbeat statistics for this connection.
        """

        return copy.copy(self._heartbeat_stats)

    async def _read_single_frame(self) -> Frame:
        """
        Reads a single frame from the AMQP connection.
        """

        frame = self._parser.next_frame()
        if frame is not NEED_DATA:
            await checkpoint()
            return frame

        while True:
            data = await self._sock.receive(4096)
            self._parser.receive_data(data)

            frame = self._parser.next_frame()
            if frame is not NEED_DATA:
                return frame

    # bleh, this is just a hotfix for data contention. i could probably fix this by driving the
    # writer with a background task writer, but that seems like it would achieve the exact same
    # thing...?
    async def _send(self, data: bytes):
        async with self._write_lock:
            await self._sock.send(data)

    async def _send_method_frame(self, channel: int, payload: MethodPayload):
        """
        Sends a single method frame.
        """

        data = self._parser.write_method_frame(channel, payload)
        await self._send(data)

    async def _send_header_frame(
        self,
        channel: int,
        *,
        method_klass: ClassID,
        body_length: int,
        headers: BasicHeader,
    ):
        """
        Sends a single header frame.
        """

        data = self._parser.write_header_frame(channel, method_klass, body_length, headers)
        await self._send(data)

    async def _send_body_frames(self, channel: int, body: bytes):
        """
        Sends multiple body frames.
        """

        data = self._parser.write_body_frames(channel, body, max_frame_size=self._max_frame_size)

        if not data:
            await checkpoint()
            return

        for frame in data:
            await self._send(frame)

    async def _close_ungracefully(self):
        """
        Closes the connection ungracefully.
        """

        if self._closed:
            return await checkpoint()

        try:
            await self._sock.aclose()
        finally:
            self._closed = True

    async def _do_startup_handshake(self, username: str, password: str, vhost: str):
        """
        Does the startup handshake.
        """

        state = AMQPState.INITIAL
        logger.debug("Sending AMQP handshake...")

        open_message = b"AMQP\x00\x00\x09\x01"
        await self._send(open_message)

        while self.open:
            incoming_frame = await self._read_single_frame()
            # this can *never* reasonably happen during the handshake
            assert isinstance(incoming_frame, MethodFrame), "incoming frame was not a method???"

            if isinstance(incoming_frame.payload, ConnectionClosePayload):
                close_ok = ConnectionCloseOkPayload()
                try:
                    await self._send_method_frame(0, close_ok)
                finally:
                    raise UnexpectedCloseError.of(incoming_frame.payload)

            if state == AMQPState.INITIAL:
                payload = incoming_frame.payload

                if not isinstance(payload, ConnectionStartPayload):  # pragma: no cover
                    await self._close_ungracefully()
                    raise InvalidPayloadTypeError(ConnectionStartPayload, payload)

                version = (payload.version_major, payload.version_minor)

                if version != (0, 9):  # pragma: no cover
                    await self._close_ungracefully()
                    raise InvalidProtocolError(f"Expected AMQP 0-9-1, but server speaks {version}")

                mechanisms = set(payload.mechanisms.decode(encoding="utf-8").split(" "))
                locales = set(payload.mechanisms.decode(encoding="utf-8").split(" "))

                platform = payload.properties["platform"].decode(encoding="utf-8")
                product = payload.properties["product"].decode(encoding="utf-8")
                if product == "RabbitMQ":  # pragma: no cover
                    logger.debug("Enabling RabbitMQ-specific code paths...")
                    self._is_rabbitmq = True

                version = payload.properties["version"].decode(encoding="utf-8")
                logger.debug(f"Connected to {product} v{version} ({platform})")

                caps: Dict[str, bool] = payload.properties["capabilities"]

                # this is the only capability we require
                if not caps.get("publisher_confirms", False):  # pragma: no cover
                    raise AMQPError("Server does not support publisher confirms, which we require")

                self._server_capabilites = caps
                for cap, value in self._server_capabilites.items():
                    if value:
                        logger.trace(f"Server supports capability {cap}")

                if "PLAIN" not in mechanisms:  # pragma: no cover
                    # we only speak plain (for now...)
                    await self._close_ungracefully()
                    raise AMQPStateError(
                        f"Expected PLAIN authentication method, but we only have {mechanisms}"
                    )

                sasl_response = b"\x00%s\x00%s" % (
                    username.encode("utf-8"),
                    password.encode("utf-8"),
                )
                ok_frame = ConnectionStartOkPayload(
                    properties=self.get_client_properties(),
                    mechanism="PLAIN",
                    response=sasl_response,
                    locale="en_US",
                )
                await self._send_method_frame(0, ok_frame)
                state = AMQPState.RECEIVED_START

            elif state == AMQPState.RECEIVED_START:
                payload = incoming_frame.payload
                if isinstance(payload, ConnectionTunePayload):
                    wanted_channel_size = min(payload.max_channels, 65535)
                    logger.debug(
                        f"Server asks for {payload.max_channels} channels, "
                        f"we're asking for {wanted_channel_size} channels"
                    )

                    self._channels = BitSet(wanted_channel_size)
                    self._channels[0] = True

                    wanted_frame_size = min(payload.max_frame_size, self._max_frame_size)
                    logger.debug(
                        f"Server asks for {payload.max_frame_size}B frame sizes, "
                        f"we're asking for {wanted_frame_size}B frame sizes"
                    )
                    self._max_frame_size = wanted_frame_size

                    hb_interval = min(payload.heartbeat_delay, self._heartbeat_interval)
                    logger.debug(
                        f"Server asks for {payload.heartbeat_delay} seconds between "
                        f"heartbeats, we're asking for {hb_interval} seconds"
                    )
                    self._actual_heartbeat_interval = hb_interval

                    tune_ok = ConnectionTuneOkPayload(
                        max_channels=wanted_channel_size,
                        max_frame_size=wanted_frame_size,
                        heartbeat_delay=hb_interval,
                    )
                    await self._send_method_frame(0, tune_ok)

                    # open the connection now
                    open = ConnectionOpenPayload(virtual_host=vhost)
                    await self._send_method_frame(0, open)
                    state = AMQPState.RECEIVED_TUNE
                else:  # pragma: no cover
                    raise InvalidPayloadTypeError(ConnectionTunePayload, payload)

            elif state == AMQPState.RECEIVED_TUNE:
                payload = incoming_frame.payload
                if isinstance(payload, ConnectionOpenOkPayload):
                    # we are open
                    logger.info("AMQP connection is ready to go")
                    break
                else:  # pragma: no cover
                    raise InvalidPayloadTypeError(ConnectionOpenOkPayload, payload)

    async def _open_channel(self) -> Channel:
        """
        Opens a new channel.
        """

        for idx, is_used in enumerate(self._channels):
            if not is_used:
                self._channels[idx] = True
                break
        else:
            # todo better error
            raise RuntimeError("All channel IDs have been used")

        logger.debug(f"Opening new AMQP channel with {idx=}")

        channel_object = Channel(idx, self, self._channel_buffer_size)
        self._channel_channels[idx] = channel_object

        try:
            async with anyio.create_task_group() as tg:
                open = ChannelOpenPayload()

                tg.start_soon(partial(self._send_method_frame, idx, open))
                await channel_object._wait_until_open()
        except:
            self._channel_channels.pop(idx)
            raise

        logger.debug(f"Enabling channel QOS with {self._channel_buffer_size // 3} max payloads.")

        # enable qos and select
        qos = BasicQOSPayload(
            prefetch_size=0,
            prefetch_count=self._channel_buffer_size // 3,  # METHOD, HEADER, BODY
            global_=False,
        )
        await channel_object._send_and_receive_frame(qos, BasicQOSOkPayload)

        select = ConfirmSelectPayload(no_wait=False)
        await channel_object._send_and_receive_frame(select, ConfirmSelectOkPayload)

        return channel_object

    async def _close_channel(self, id: int):
        """
        Closes a channel.
        """

        if self._server_requested_close:
            logger.warning(f"Server requested close without closing channel #{id}")
            # TODO: Should this be a cancel shielded checkpoint?
            return

        frame = ChannelClosePayload(
            reply_code=ReplyCode.success, reply_text="Normal close", class_id=0, method_id=0
        )
        await self._send_method_frame(id, frame)

    async def _handle_control_frame(self, frame: MethodFrame):
        """
        Handles a control frame.
        """

        payload = frame.payload

        if isinstance(payload, ConnectionClosePayload):
            # server closing connection
            logger.info("Server requested close...")
            self._server_requested_close = True
            self._close_info = payload

            # cancel our worker tasks if needed
            if self._cancel_scope is not None:
                # noinspection PyAsyncCall
                self._cancel_scope.cancel()

    async def _remove_channel(self, channel_id: int, payload: Optional[ChannelClosePayload]):
        """
        Removes a channel.
        """

        self._channels[channel_id] = False
        chan = self._channel_channels.pop(channel_id)
        await chan._close(payload)

    async def _enqueue_frame(self, channel: Channel, frame: Frame):
        """
        Enqueues a frame onto the channel object.

        :param channel: The channel to handle.
        :param frame: The frame to handle.
        """

        # pfft, there used to be a whole doc comment about enabling/disabling flows but turns out
        # i can't read and there's just a qos() method.

        threshold = channel.max_buffer_size - 1
        count = threshold - channel.current_buffer_size
        # simple case
        if count >= 1:
            logger.debug(
                f"Enqueueing frame on channel {channel.id}, {count} frames left in the buffer"
            )
            return await channel._enqueue_delivery(frame)

        # complex case
        logger.critical(
            f"!!! CHANNEL {channel.id} IS ABOUT TO BLOCK: This WILL cause a deadlock !!!"
        )
        logger.critical("The most likely cause is too large messages!")
        logger.critical(
            "You *must* adjust your desired frame size on both the client and the "
            "server, raise the channel buffer, or have faster consumers."
        )
        logger.critical(
            f"Current frame size is {self._max_frame_size}, and the current buffer "
            f"size is {self._channel_buffer_size}."
        )
        logger.debug(f"Enqueuing frame on blocked channel {channel.id}")
        return await channel._enqueue_delivery(frame)

    async def _heartbeat_loop(self):
        """
        Sends heartbeats to the AMQP server.
        """

        if self._heartbeat_interval == 0:
            return

        while True:
            # heartbeat frame
            await self._send(b"\x08\x00\x00\x00\x00\x00\x00\xCE")
            await sleep(self._heartbeat_interval / 2)

    async def _listen_for_messages(self):
        """
        Listens for messages infinitely. This is the primary driving loop of the connection.
        """

        while not (self._closed or self._server_requested_close):
            try:
                with anyio.fail_after(self._actual_heartbeat_interval):
                    frame = await self._read_single_frame()
            except TimeoutError:
                logger.error(
                    f"Server failed to send any messages in {self._actual_heartbeat_interval} "
                    "seconds, disconnecting!"
                )

                await self._close_ungracefully()
                raise

            if frame.type == FrameType.HEARTBEAT:
                self._heartbeat_stats.update()
                interval = self._heartbeat_stats.interval
                if interval is None:
                    logger.debug("Received heartbeat frame")
                else:
                    logger.debug(
                        f"Received heartbeat frame (interval: {interval / 1_000_000_000:.2f}s)"
                    )

            elif frame.type == FrameType.METHOD:
                assert isinstance(frame, MethodFrame)

                # intercept control frames, e.g. close payloads
                channel = frame.channel_id
                if channel == 0:
                    await self._handle_control_frame(frame)  # type: ignore
                    continue

                # we intercept certain control frames
                # channelcloseok is handled here because channels can't deal with their own closure
                # (as they don't know when they finish)
                channel_object = self._channel_channels[channel]

                if isinstance(frame.payload, (ChannelClosePayload, ChannelCloseOkPayload)):
                    is_unclean = isinstance(frame.payload, ChannelClosePayload)
                    logger.debug(f"Channel close received, {is_unclean=}")

                    # ack the close
                    # todo: should his be here?
                    try:
                        await self._remove_channel(channel, frame.payload)
                    finally:
                        if is_unclean:
                            await self._send_method_frame(channel, ChannelCloseOkPayload())

                    continue

                elif isinstance(
                    frame.payload, (BasicDeliverPayload, BasicGetOkPayload, BasicGetEmptyPayload)
                ):
                    # requires special logic
                    await self._enqueue_frame(channel_object, frame)

                else:
                    # just delivery normally
                    try:
                        channel_object._enqueue_regular(frame)
                    except WouldBlock:
                        logger.warning(
                            f"Channel #{channel} was not listening for frame "
                            f"{method_payload_name(frame.payload)}, dropping"
                        )

            elif frame.type == FrameType.HEADER or frame.type == FrameType.BODY:
                channel = frame.channel_id
                assert channel != 0, "header frame cannot happen on control channel"
                channel_object = self._channel_channels[channel]
                await self._enqueue_frame(channel_object, frame)

    def _start_tasks(self, nursery: TaskGroup):
        """
        Starts the background tasks for this connection.
        """
        self._cancel_scope = nursery.cancel_scope
        nursery.start_soon(self._listen_for_messages)
        nursery.start_soon(self._heartbeat_loop)

    async def _close(self, reply_code: int = 200, reply_text: str = "Normal close"):
        if self._closed:
            await checkpoint()
            return

        # at this point we're being called in the finally, so we have exclusive control of the
        # socket
        # we may be being cancelled, however, so we need to shield ourselves when sending more
        # things

        with CancelScope(shield=True):
            # send a CloseOk method if the server requested our closure
            if self._server_requested_close:
                logger.debug("Acknowledging server close...")
                payload = ConnectionCloseOkPayload()
                await self._send_method_frame(0, payload)
                await self._close_ungracefully()

                if self._close_info.reply_code != 200:
                    raise UnexpectedCloseError.of(self._close_info)

            else:
                logger.debug("Sending close payload...")

                payload = ConnectionClosePayload(
                    reply_code=ReplyCode(reply_code), reply_text=reply_text, class_id=0, method_id=0
                )

                await self._send_method_frame(0, payload)

                try:
                    while True:
                        # drain previous frames
                        # channels will send their own close frame before we get a chance to drain
                        # it

                        try:
                            reply = await self._read_single_frame()
                        except EndOfStream:
                            raise AMQPStateError(
                                f"Expected CloseOk, but connection failed to send it"
                            )

                        if isinstance(reply, MethodFrame) and isinstance(
                            reply.payload, ConnectionCloseOkPayload
                        ):
                            logger.debug("Received CloseOk, closing connection")
                            return
                finally:
                    await self._close_ungracefully()

    def open_channel(self) -> AsyncContextManager[Channel]:
        """
        Opens a new channel.

        :return: An asynchronous context manager that will open a new channel.
        """

        if not self.open:
            raise RuntimeError("The connection is not open")

        @asynccontextmanager
        async def cm():
            channel = await self._open_channel()
            try:
                yield channel
            finally:
                # don't try and re-close the channel if the channel was closed server-side

                if not channel._closed:
                    channel._closed = True

                    with anyio.CancelScope(shield=True):
                        await self._close_channel(channel.id)

        return cm()

    async def close(self, reply_code: int = 200, reply_text: str = "Normal close"):
        """
        Closes the connection. This method is idempotent.

        There's no real reason to call this method.

        :param reply_code: The code to send when closing.
        :param reply_text: The text to send when replying.
        :return: Nothing.
        """

        # close any background tasks
        if not self._cancel_scope.cancel_called:
            # noinspection PyAsyncCall
            self._cancel_scope.cancel()

        await self._close(reply_code, reply_text)


async def _open_connection(
    address: Union[str, PathLike],
    *,
    port: int = 6379,
    username: str = "guest",
    password: str = "guest",
    vhost: str = "/",
    ssl_context: SSLContext = None,
    **kwargs,
) -> AMQPConnection:
    """
    Actually implements opening the connection and performing the startup handshake.
    """

    if isinstance(address, os.PathLike) or address.startswith("/"):
        path = os.fspath(address)
        logger.debug(f"Opening connection to {path}")
        sock = await anyio.connect_unix(path)
    else:
        logger.debug(f"Opening TCP connection to {address}:{port}")
        sock = await anyio.connect_tcp(
            address,
            remote_port=port,
            tls=ssl_context is not None,
            ssl_context=ssl_context,
            tls_standard_compatible=True,
        )

    connection = AMQPConnection(sock, **kwargs)
    # noinspection PyProtectedMember
    await connection._do_startup_handshake(username, password, vhost)
    return connection


# noinspection PyIncorrectDocstring
def open_connection(
    address: Union[str, PathLike],
    *,
    port: int = 5672,
    username: str = "guest",
    password: str = "guest",
    virtual_host: str = "/",
    ssl_context: SSLContext = None,
    **kwargs,
) -> AsyncContextManager[AMQPConnection]:
    """
    Opens a new connection to the AMQP 0-9-1 server. This is an asynchronous context manager.

    Required parameters:

    :param address: The address of the server or the *absolute path* of its Unix socket.

    Optional parameters:

    :param port: The port to connect to. Ignores for Unix sockets. Defaults to 5672.
    :param username: The username to connect using. Defaults to ``guest``.
    :param password: The password to authenticate with. Defaults to ``guest``.
    :param virtual_host: The AMQP virtual host to connect to. Defaults to ``/``.
    :param ssl_context: The SSL context to connect with. Defaults to None.

    In addition, some parameters are passed directly to the class, which allows customising
    some protocol details.

    :param heartbeat_interval: The heartbeat interval to negotiate with, in seconds. This may not
                               be the actual heartbeat interval used.
    :param channel_buffer_size: The buffer size for channel messages.
    :param desired_frame_size: The maximum body message frame size desired.
    """

    @asynccontextmanager
    async def _do_open():
        conn = await _open_connection(
            address=address,
            port=port,
            username=username,
            password=password,
            vhost=virtual_host,
            ssl_context=ssl_context,
            **kwargs,
        )

        try:
            async with anyio.create_task_group() as nursery:
                conn._start_tasks(nursery)
                try:
                    yield conn
                except BaseException as e:
                    logger.error(f"Caught error {type(e).__qualname__}, closing connection")
                    raise
                finally:
                    # noinspection PyAsyncCall
                    nursery.cancel_scope.cancel()
        finally:
            await conn.close()

    return _do_open()



from typing import Any, Union, get_args, get_origin

from attr import Attribute

from serena.utils.buffer import DecodingBuffer, EncodingBuffer


def aq_type(name):
    return {"amqp_type": name}


def encode_attrs_attribute(buf: EncodingBuffer, att: Attribute, value: Any):
    """
    Encodes an attrs attribute on a class.

    :param buf: The buffer to encode into.
    :param att: The attribute to encode.
    :param value: The value of the field.
    :return:
    """

    field_type = get_origin(att.type) or att.type

    if field_type is Union:
        unwrapped_field_type = get_args(att.type)[0]
        field_type = get_origin(unwrapped_field_type) or unwrapped_field_type

    if field_type is str:
        fn = buf.write_short_string

    elif field_type is bytes:
        fn = buf.write_long_string

    elif field_type is bool:
        fn = buf.write_bit

    elif field_type is dict:
        fn = buf.write_table

    elif field_type is list:
        # todo
        raise NotImplementedError("list types")

    else:
        type_ = att.metadata.get("amqp_type")
        if not type_:  # not an amqp field, probably default field?
            return

        fn = getattr(buf, f"write_{type_}")

    fn(value)


def decode_attrs_attribute(buf: DecodingBuffer, att: Attribute) -> Any:
    """
    Decodes an attrs field on a class.

    :param buf: The buffer to read from.
    :param att: The attribute to decode.
    :return: The decoded value, or None if there was nothing to decode.
    """

    field_type = get_origin(att.type) or att.type

    # we only allow Optional unions so this is always safe
    if field_type is Union:
        unwrapped_field_type = get_args(att.type)[0]
        field_type = get_origin(unwrapped_field_type) or unwrapped_field_type

    if field_type is str:
        fn = buf.read_short_string

    elif field_type is bytes:
        fn = buf.read_long_string

    elif field_type is bool:
        fn = buf.read_bit

    elif field_type is dict:
        fn = buf.read_table

    elif field_type is list:
        fn = buf.read_array

    else:
        type_ = att.metadata.get("amqp_type")
        if not type_:  # not an amqp field, probably default field?
            return None

        fn = getattr(buf, f"read_{type_}")

    return fn()



from __future__ import annotations

import struct
from contextlib import contextmanager
from datetime import datetime
from io import BytesIO
from typing import Any, ContextManager, Dict, List, overload


class DecodingBuffer(object):
    """
    A buffer that allows automatic decoding of AMQP wire protocol objects.
    """

    def __init__(self, payload_data: bytes):
        """
        :param payload_data: The payload itself to decode.
        """

        self._data = payload_data
        self._offset = 0

        # copied when doing a read bit
        self._last_bit_data = 0
        self._last_bit_offset = 0

    @property
    def has_data(self) -> bool:
        """
        Returns True if there is still data to read.
        """

        return self._offset < len(self._data)

    def _unpack(self, fmt):
        fmt = "!" + fmt
        size = struct.calcsize(fmt)
        items = struct.unpack_from(fmt, self._data, self._offset)
        self._offset += size

        # wipe bit data
        self._last_bit_data = 0
        self._last_bit_offset = 0

        return items

    def _read_size(self, size: int) -> bytes:
        data = self._data[self._offset : self._offset + size]
        self._offset += size

        self._last_bit_data = 0
        self._last_bit_offset = 0

        return data

    def read_octet_signed(self) -> int:
        """
        Reads a single signed octet from the stream.
        """

        return self._unpack("b")[0]

    def read_octet(self) -> int:
        """
        Reads a single octet from the stream.
        """

        return self._unpack("B")[0]

    def read_short_signed(self) -> int:
        """
        Reads a single signed short from the stream.
        """

        return self._unpack("h")[0]

    def read_short(self) -> int:
        """
        Reads a single short (2-byte) from the stream.
        """

        return self._unpack("H")[0]

    def read_long_signed(self) -> int:
        """
        Reads a single signed long (4-byte) from the stream.
        """

        return self._unpack("i")[0]

    def read_long(self) -> int:
        """
        Reads a single long (4-byte) from the stream.
        """

        return self._unpack("I")[0]

    def read_longlong_signed(self) -> int:
        """
        Reads a single signed long long (8-byte) from the stream.
        """

        return self._unpack("q")[0]

    def read_longlong(self) -> int:
        """
        Reads a single long-long (8-byte) from the stream.
        """

        return self._unpack("Q")[0]

    def read_short_string(self) -> str:
        """
        Reads a single short string from the stream.
        """

        size = self.read_octet()
        return self._read_size(size).decode(encoding="utf-8")

    def read_long_string(self) -> bytes:
        """
        Reads a single long string from the stream.
        """

        size = self.read_long()
        return self._read_size(size)

    def read_field_value(self) -> Any:
        """
        Reads a type-prefixed field value.
        """

        buf = self
        type = chr(self.read_octet())

        if type == "t":
            item = buf.read_octet() == 1
        elif type == "b":
            item = buf.read_octet_signed()
        elif type == "B":
            item = buf.read_octet()
        elif type == "U":
            item = buf.read_short_signed()
        elif type == "u":
            item = buf.read_short()
        elif type == "I":
            item = buf.read_long_signed()
        elif type == "i":
            item = buf.read_long()
        elif type == "L":
            item = buf.read_longlong_signed()
        elif type == "l":
            # item = buf.read_longlong()  # see errata
            item = buf.read_longlong_signed()
        elif type == "f" or type == "d":
            item = buf._unpack(type)
        elif type == "D":
            raise ValueError("Fuck you")
        elif type == "s":
            # item = buf.read_short_string()  # see errata
            item = buf.read_short_signed()
        elif type == "S" or type == "x":
            item = buf.read_long_string()
        elif type == "A":
            item = self.read_array()
        elif type == "T":
            item = self.read_longlong()
        elif type == "F":
            item = self.read_table()
        elif type == "V":
            item = None
        else:
            raise ValueError(f"Unknown type code {type}")

        return item

    def read_array(self) -> List[Any]:
        """
        Reads an array of values.
        """

        item_count = self.read_long()
        return [self.read_field_value() for _ in range(0, item_count)]

    def read_table(self) -> Dict[str, Any]:
        """
        Reads a table from the stream.
        """

        table = self.read_long_string()
        buf = DecodingBuffer(table)

        result = {}

        while buf.has_data:
            key = buf.read_short_string()
            item = buf.read_field_value()
            result[key] = item

        return result

    def read_bit(self) -> bool:
        """
        Reads a single bit from the stream, as a boolean.
        """

        if 0 < self._last_bit_offset < 8:
            bit = ((self._last_bit_data) & (1 << self._last_bit_offset)) == 1
            self._last_bit_offset += 1
            return bit

        # copy next bit off
        data = self._unpack("B")[0]
        self._last_bit_data = data
        bit = (data & 1) == 1
        self._last_bit_offset += 1
        return bit


class EncodingBuffer(object):
    """
    A buffer that writes data in AMQP format.
    """

    def __init__(self):
        self._table_mode = False
        self._data = BytesIO()

        self._last_bit_data = 0
        self._last_bit_offset = 0

    def _write(self, data: bytes):
        if self._last_bit_offset > 0:
            self._data.write(self._last_bit_data.to_bytes(length=1, byteorder="big"))
            self._last_bit_data = 0
            self._last_bit_offset = 0

        self._data.write(data)

    def get_data(self) -> bytes:
        """
        Gets the raw data in this buffer. This preserves the previous cursor, so data can be
        written even after calling this method.
        """

        pos = self._data.tell()
        self._data.seek(0)
        data = self._data.read()
        self._data.seek(pos)

        return data

    def write_long_string(self, data: bytes):
        """
        Writes a long string to the buffer.
        """

        if self._table_mode:
            self._write(b"S")

        size = struct.pack(">I", len(data))
        self._write(size)
        self._write(data)

    def _write_string(self, data: str):
        encoded = data.encode("utf-8")
        size = struct.pack(">B", len(encoded))
        self._write(size)
        self._write(encoded)

    def write_short_string(self, data: str):
        """
        Writes a short string to the buffer.
        """

        if self._table_mode:
            raise ValueError("short strings have no table type")

        self._write_string(data)

    def write_octet(self, value: int):
        """
        Writes a single byte to the buffer.
        """

        if self._table_mode:
            self._write(b"B")

        self._write(struct.pack(">B", value))

    def write_octet_signed(self, value: int):
        """
        Writes a single signed byte to the buffer.
        """

        if self._table_mode:
            self._write(b"B")

        self._write(struct.pack(">b", value))

    def write_short(self, value: int):
        """
        Writes a single short to the buffer.
        """

        if self._table_mode:
            self._write(b"u")

        self._write(struct.pack(">H", value))

    def write_short_signed(self, value: int):
        """
        Writes a single signed short to the buffer.
        """

        if self._table_mode:
            self._write(b"U")

        self._write(struct.pack(">h", value))

    def write_long(self, value: int):
        """
        Writes a single long to the buffer.
        """

        if self._table_mode:
            self._write(b"i")

        self._write(struct.pack(">I", value))

    def write_long_signed(self, value: int):
        """
        Writes a single signed long to the buffer.
        """

        if self._table_mode:
            self._write(b"I")

        self._write(struct.pack(">i", value))

    def write_longlong(self, value: int):
        """
        Writes a single long long to the buffer.
        """

        if self._table_mode:
            raise ValueError("unsigned longlongs have no table type")

        self._write(struct.pack(">Q", value))

    def write_longlong_signed(self, value: int):
        """
        Writes a single signed long long to the buffer.
        """

        if self._table_mode:
            self._write(b"l")

        self._write(struct.pack(">q", value))

    @overload
    def write_timestamp(self, value: datetime):
        ...

    @overload
    def write_timestamp(self, value: int):
        ...

    def write_timestamp(self, value):
        """
        Writes a timestamp to the buffer.
        """

        if self._table_mode:
            self._write(b"T")

        if isinstance(value, datetime):
            value = int(value.timestamp())

        # time_t, long
        self._write(struct.pack(">L", value))

    def write_float(self, value: float):
        """
        Writes a single precision float to the buffer.
        """

        if self._table_mode:
            self._write(b"f")

        self._write(struct.pack(">f", value))

    def write_double(self, value: float):
        """
        Writes a double precision float to the buffer.
        """

        if self._table_mode:
            self._write(b"d")

        self._write(struct.pack(">d", value))

    def write_bit(self, value: bool):
        """
        Writes a single bit to the buffer.
        """

        if self._table_mode:
            self._write(b"t")
            self._write(struct.pack(">B", value))
        else:
            if self._last_bit_offset >= 8:
                self._write(b"")  # forces a bit write
            else:
                self._last_bit_data = (self._last_bit_data << 1) | value
                self._last_bit_offset += 1

    def force_write_bits(self):
        """
        Forces a trailing bit write if needed.
        """

        self._write(b"")

    @contextmanager
    def _table_cm(self):
        buf = TableWriter()
        yield buf

        if self._table_mode:
            self._write(b"F")

        # write long string manually, as `write_long_string` writes an extraneous long string
        # field value if this is a nested table
        data = buf.get_data()
        size = struct.pack(">I", len(data))
        self._write(size)
        self._write(data)

    def start_writing_table(self) -> ContextManager[TableWriter]:
        """
        Writes a table to the stream. This is a context manager that allows writing key/values to
        a table.
        """

        return self._table_cm()

    def write_table(self, table: Dict[str, Any]):
        """
        Writes a complete table.
        """

        with self.start_writing_table() as writer:
            for key, value in table.items():
                writer.automatically_write_value(key, value)


class TableWriter(EncodingBuffer):
    """
    A subclass of the encoding buffer that has extensions for table methods.
    """

    def __init__(self):
        super().__init__()

        self._table_mode = True

    def write_key(self, key: str):
        """
        Writes a standalone key to the buffer. This ignores table mode.
        """

        self._write_string(key)

    def automatically_write_value(self, key: str, value: Any):
        """
        Automatically writes a value. This will detect the type of the value and write the
        appropriate type. This is primarily useful for writing table data.

        :param key: The key for this value.
        :param value: The value to write.
        """

        self.write_key(key)

        if isinstance(value, bool):
            self.write_bit(value)
            self.force_write_bits()

        elif isinstance(value, int):
            bl = value.bit_length()
            if bl <= 32:
                self.write_long(value)
            else:
                self.write_longlong(value)

        elif isinstance(value, float):
            self.write_double(value)

        elif isinstance(value, str):
            self.write_long_string(value.encode("utf-8"))

        elif isinstance(value, bytes):
            self.write_long_string(value)

        elif isinstance(value, datetime):
            self.write_timestamp(value)

        elif isinstance(value, dict):
            self.write_table(value)

        else:
            raise ValueError(f"Unknown item: {value} ({type(value)})")



from __future__ import annotations

import abc
from typing import Any, ClassVar, Dict, Generic, TypeVar

import attr

from serena.enums import ClassID, ReplyCode
from serena.frame import Frame, FrameType
from serena.payloads.encoding import (
    aq_type,
    decode_attrs_attribute,
    encode_attrs_attribute,
)
from serena.utils.buffer import DecodingBuffer, EncodingBuffer


@attr.s(frozen=True, slots=True, init=True)
class MethodPayload(abc.ABC):
    """
    Base class for all method payloads.
    """

    #: The class ID for this payload.
    klass: ClassVar[ClassID]

    #: The method ID for this payload. These are shared between classes and are not unique.
    method: ClassVar[int]

    #: If this is a payload sent FROM the client. This may be True on payloads received by the
    #: client, too.
    is_client_side: ClassVar[bool]


_PAYLOAD_TYPE = TypeVar("_PAYLOAD_TYPE", bound=MethodPayload)


@attr.s(frozen=True, slots=True)
class MethodFrame(Frame, Generic[_PAYLOAD_TYPE]):
    """
    A frame that carries a method body.
    """

    type = FrameType.METHOD

    #: The payload for this frame.
    payload: _PAYLOAD_TYPE = attr.ib()


## == CONNECTION == ##
@attr.s(frozen=True, slots=True)
class ConnectionStartPayload(MethodPayload):
    """
    Payload for the ``start`` method.
    """

    klass = ClassID.CONNECTION
    method = 10
    is_client_side = False

    #: The major version of the protocol.
    version_major: int = attr.ib(metadata=aq_type("octet"))

    #: The minor version of the protocol.
    version_minor: int = attr.ib(metadata=aq_type("octet"))

    #: The server properties.
    properties: Dict[str, Any] = attr.ib()

    #: The available security mechanisms.
    mechanisms: bytes = attr.ib()

    #: The available message locales.
    locales: bytes = attr.ib()


@attr.s(frozen=True, slots=True)
class ConnectionStartOkPayload(MethodPayload):
    """
    Payload for the ``start-ok`` method.
    """

    klass = ClassID.CONNECTION
    method = 11
    is_client_side = True

    #: The client properties.
    properties: Dict[str, Any] = attr.ib()

    #: The selected security mechanism.
    mechanism: str = attr.ib()

    #: The security response data.
    response: bytes = attr.ib()

    #: The selected locale.
    locale: str = attr.ib()


@attr.s(frozen=True, slots=True)
class ConnectionSecurePayload(MethodPayload):
    """
    Payload for the ``secure`` method.
    """

    klass = ClassID.CONNECTION
    method = 20
    is_client_side = False

    #: The security challenge data.
    challenge: bytes = attr.ib()


@attr.s(frozen=True, slots=True)
class ConnectionSecureOkPayload(MethodPayload):
    """
    Payload for the ``secure-ok`` method.
    """

    klass = ClassID.CONNECTION
    method = 21
    is_client_side = True

    #: The security response data.
    response: bytes = attr.ib()


@attr.s(frozen=True, slots=True)
class ConnectionTunePayload(MethodPayload):
    """
    Payload for the ``tune`` method.
    """

    klass = ClassID.CONNECTION
    method = 30
    is_client_side = False

    #: The server's proposed maximum channels.
    max_channels: int = attr.ib(metadata=aq_type("short"))

    #: The server's proposed maximum frame size.
    max_frame_size: int = attr.ib(metadata=aq_type("long"))

    #: The server's desired heartbeat delay.
    heartbeat_delay: int = attr.ib(metadata=aq_type("short"))


@attr.s(frozen=True, slots=True)
class ConnectionTuneOkPayload(MethodPayload):
    """
    Payload for the ``tune-ok`` method.
    """

    klass = ClassID.CONNECTION
    method = 31
    is_client_side = True

    #: The client's negotiated maximum channels.
    max_channels: int = attr.ib(metadata=aq_type("short"))

    #: The client's negotiated maximum frame size.
    max_frame_size: int = attr.ib(metadata=aq_type("long"))

    #: The client's desired heartbeat delay.
    heartbeat_delay: int = attr.ib(metadata=aq_type("short"))


@attr.s(frozen=True, slots=True)
class ConnectionOpenPayload(MethodPayload):
    """
    Payload for the ``open`` method.
    """

    klass = ClassID.CONNECTION
    method = 40
    is_client_side = False

    #: The virtual host to open a connection to.
    virtual_host: str = attr.ib()

    reserved_1: str = attr.ib(default="")
    reserved_2: bool = attr.ib(default=True)


@attr.s(frozen=True, slots=True)
class ConnectionOpenOkPayload(MethodPayload):
    """
    Payload for the ``open-ok`` method.
    """

    klass = ClassID.CONNECTION
    method = 41
    is_client_side = True

    reserved_1: str = attr.ib()


@attr.s(frozen=True, slots=True)
class ConnectionClosePayload(MethodPayload):
    """
    Payload for the ``close`` method.
    """

    klass = ClassID.CONNECTION
    method = 50
    is_client_side = True

    #: The code for the error that caused this close.
    reply_code: ReplyCode = attr.ib(converter=ReplyCode, metadata=aq_type("short"))

    #: The text for the error that caused this close.
    reply_text: str = attr.ib()

    #: The class of the method that caused this close.
    class_id: int = attr.ib(metadata=aq_type("short"))

    #: The class of the method that caused this close.
    method_id: int = attr.ib(metadata=aq_type("short"))


@attr.s(frozen=True, slots=True)
class ConnectionCloseOkPayload(MethodPayload):
    """
    Payload (empty) for the ``close-ok`` method.
    """

    klass = ClassID.CONNECTION
    method = 51
    is_client_side = True


## == CHANNEL == ##
@attr.s(frozen=True, slots=True)
class ChannelOpenPayload(MethodPayload):
    """
    Payload for the ``open`` method.
    """

    klass = ClassID.CHANNEL
    method = 10
    is_client_side = False

    reserved_1: str = attr.ib(default="")


@attr.s(frozen=True, slots=True)
class ChannelOpenOkPayload(MethodPayload):
    """
    Payload for the ``open-ok`` method.
    """

    klass = ClassID.CHANNEL
    method = 11
    is_client_side = True

    reserved_1: bytes = attr.ib()


@attr.s(frozen=True, slots=True)
class ChannelFlowPayload(MethodPayload):
    """
    Payload for the ``flow`` method.
    """

    klass = ClassID.CHANNEL
    method = 20
    is_client_side = True

    #: If the channel should start processing messages again or not.
    active: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class ChannelFlowOkPayload(MethodPayload):
    """
    Payload for the ``flow-ok`` method.
    """

    klass = ClassID.CHANNEL
    method = 21
    is_client_side = True

    #: See :attr:`.FlowPayload.active`.
    active: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class ChannelClosePayload(MethodPayload):
    """
    Payload for the ``close`` method.
    """

    klass = ClassID.CHANNEL
    method = 40
    is_client_side = True

    #: The code for the error that caused this close.
    reply_code: ReplyCode = attr.ib(converter=ReplyCode, metadata=aq_type("short"))

    #: The text for the error that caused this close.
    reply_text: str = attr.ib()

    #: The class of the method that caused this close.
    class_id: int = attr.ib(metadata=aq_type("short"))

    #: The class of the method that caused this close.
    method_id: int = attr.ib(metadata=aq_type("short"))


@attr.s(frozen=True, slots=True)
class ChannelCloseOkPayload(MethodPayload):
    """
    Payload for the ``close-ok`` method.
    """

    klass = ClassID.CHANNEL
    method = 41
    is_client_side = True

    # empty body


## EXCHANGE ##
@attr.s(frozen=True, slots=True)
class ExchangeDeclarePayload(MethodPayload):
    """
    Payload for the ``declare`` method.
    """

    klass = ClassID.EXCHANGE
    method = 10
    is_client_side = False

    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the exchange to be created.
    name: str = attr.ib()

    #: The type of the exchange to be created.
    exchange_type: str = attr.ib()

    #: If True, no Declare-Ok method will be sent by the server.
    no_wait: bool = attr.ib()

    #: If True, the exchange may not bee used ddirectly by publishers, but only when bound to other
    #: exchanges.
    internal: bool = attr.ib()

    #: If True, the exchange will be deleted when all queues have finished using it.
    auto_delete: bool = attr.ib()

    #: If True, the the exchange will survive server restarts.
    durable: bool = attr.ib()

    #: If True, the server will return a DeclareOk if the exchange exists, and an error if it
    #: doesn't.
    passive: bool = attr.ib()

    #: Implementation-specific arguments for the declaration.
    arguments: Dict[str, Any] = attr.ib()


@attr.s(frozen=True, slots=True)
class ExchangeDeclareOkPayload(MethodPayload):
    """
    Payload for the ``declare-ok`` method.
    """

    klass = ClassID.EXCHANGE
    method = 11
    is_client_side = True

    # empty body


@attr.s(frozen=True, slots=True)
class ExchangeDeletePayload(MethodPayload):
    """
    Payload for the ``delete`` method.
    """

    klass = ClassID.EXCHANGE
    method = 20
    is_client_side = False

    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the exchange to delete.
    name: str = attr.ib()

    #: If True, no Declare-Ok method will be sent by the server.
    no_wait: bool = attr.ib()

    #: If True, the server will only delete the exchange if there are no queue bindings.
    if_unused: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class ExchangeDeleteOkPayload(MethodPayload):
    """
    Payload for the ``delete-ok`` method.
    """

    klass = ClassID.EXCHANGE
    method = 21
    is_client_side = True

    # empty body


@attr.s(frozen=True, slots=True)
class ExchangeBindPayload(MethodPayload):
    """
    Payload for the ``bind`` method.
    """

    klass = ClassID.EXCHANGE
    method = 30
    is_client_side = False

    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the destination exchange to bind.
    destination_name: str = attr.ib()

    #: The name of the source exchange to bind.
    source_name: str = attr.ib()

    #: The routing key to bind using.
    routing_key: str = attr.ib()

    #: If True, no Bind-Ok method will be sent by the server.
    no_wait: bool = attr.ib()

    #: Implementation-specific arguments for the declaration.
    arguments: Dict[str, Any] = attr.ib()


@attr.s(frozen=True, slots=True)
class ExchangeBindOkPayload(MethodPayload):
    """
    Payload for the ``bind-ok`` method.
    """

    klass = ClassID.EXCHANGE
    method = 31
    is_client_side = True

    # empty body


@attr.s(frozen=True, slots=True)
class ExchangeUnBindPayload(MethodPayload):
    """
    Payload for the ``unbind`` method.
    """

    klass = ClassID.EXCHANGE
    method = 40
    is_client_side = False

    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the destination exchange to bind.
    destination_name: str = attr.ib()

    #: The name of the source exchange to bind.
    source_name: str = attr.ib()

    #: The routing key to bind using.
    routing_key: str = attr.ib()

    #: If True, no Bind-Ok method will be sent by the server.
    no_wait: bool = attr.ib()

    #: Implementation-specific arguments for the declaration.
    arguments: Dict[str, Any] = attr.ib()


@attr.s(frozen=True, slots=True)
class ExchangeUnBindOkPayload(MethodPayload):
    """
    Payload for the ``bind-ok`` method.
    """

    klass = ClassID.EXCHANGE
    method = 41
    is_client_side = True

    # empty body


## QUEUE ##
@attr.s(frozen=True, slots=True)
class QueueDeclarePayload(MethodPayload):
    """
    Payload for the ``declare`` method.
    """

    klass = ClassID.QUEUE
    method = 10
    is_client_side = False

    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the queue being declared. May be empty.
    name: str = attr.ib()

    #: If True, no Declare-Ok method will be sent by the server.
    no_wait: bool = attr.ib()

    #: If True, then the queue is automatically deleted when all consumers are finished using it.
    auto_delete: bool = attr.ib()

    #: If True, then the queue is exclusive to this connection.
    exclusive: bool = attr.ib()

    #: If True, then the queue will persist through restarts.
    durable: bool = attr.ib()

    #: If True, the server will return a DeclareOk if the queue exists, and an error if it doesn't.
    passive: bool = attr.ib()

    #: Implementation-specific arguments for the declaration.
    arguments: Dict[str, Any] = attr.ib()


@attr.s(frozen=True, slots=True)
class QueueDeclareOkPayload(MethodPayload):
    """
    Payload for the ``declare-ok`` method.
    """

    klass = ClassID.QUEUE
    method = 11
    is_client_side = True

    #: The name of the queue.
    name: str = attr.ib()

    #: The number of the messages present in the queue.
    message_count: int = attr.ib(metadata=aq_type("long"))

    #: The number of consumers consuming from the queue.
    consumer_count: int = attr.ib(metadata=aq_type("long"))


@attr.s(frozen=True, slots=True)
class QueueBindPayload(MethodPayload):
    """
    Payload for the ``bind`` method.
    """

    klass = ClassID.QUEUE
    method = 20
    is_client_side = False

    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the queue to bind.
    queue_name: str = attr.ib()

    #: The name of the exchange to bind.
    exchange_name: str = attr.ib()

    #: The message routing key.
    routing_key: str = attr.ib()

    #: If True, no Bind-Ok method will be sent by the server.
    no_wait: bool = attr.ib()

    #: A set of implementation-specific (or exchange-specific) arguments.
    arguments: Dict[str, Any] = attr.ib()


@attr.s(frozen=True, slots=True)
class QueueBindOkPayload(MethodPayload):
    """
    Payload for the ``bind-ok`` method.
    """

    klass = ClassID.QUEUE
    method = 21
    is_client_side = True

    # empty body


@attr.s(frozen=True, slots=True)
class QueuePurgePayload(MethodPayload):
    """
    Payload for the ``purge`` method.
    """

    klass = ClassID.QUEUE
    method = 30
    is_client_side = False

    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the queue to purge.
    name: str = attr.ib()

    #: If True, no Purge-Ok method will be sent by the server.
    no_wait: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class QueuePurgeOkPayload(MethodPayload):
    """
    Payload for the ``purge-ok`` method.
    """

    klass = ClassID.QUEUE
    method = 31
    is_client_side = True

    #: The number of messages purged from the queue.
    message_count: int = attr.ib(metadata=aq_type("long"))


@attr.s(frozen=True, slots=True)
class QueueDeletePayload(MethodPayload):
    """
    Payload for the ``delete`` method.
    """

    klass = ClassID.QUEUE
    method = 40
    is_client_side = False

    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the queue to delete.
    queue_name: str = attr.ib()

    #: If True, no Delete-Ok method will be sent by the server.
    no_wait: bool = attr.ib()

    #: If True, then the queue will be deleted only if it is empty.
    if_empty: bool = attr.ib()

    #: If True, then the queue will be deleted only if it is unused.
    if_unused: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class QueueDeleteOkPayload(MethodPayload):
    """
    Payload for the ``delete-ok`` method.
    """

    klass = ClassID.QUEUE
    method = 41
    is_client_side = True

    #: The message count remaining for the queue.
    message_count: int = attr.ib(metadata=aq_type("long"))


@attr.s(frozen=True, slots=True)
class QueueUnbindPayload(MethodPayload):
    """
    Payload for the ``unbind`` method.
    """

    klass = ClassID.QUEUE
    method = 50
    is_client_side = False

    # identical to QueueBind...
    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the queue to bind.
    queue_name: str = attr.ib()

    #: The name of the exchange to bind.
    exchange_name: str = attr.ib()

    #: The message routing key.
    routing_key: str = attr.ib()

    #: If True, no Bind-Ok method will be sent by the server.
    no_wait: bool = attr.ib()

    #: A set of implementation-specific (or exchange-specific) arguments.
    arguments: Dict[str, Any] = attr.ib()


@attr.s(frozen=True, slots=True)
class QueueUnbindOkPayload(MethodPayload):
    """
    Payload for the ``unbind-ok`` method.
    """

    klass = ClassID.QUEUE
    method = 51
    is_client_side = True

    # empty body


## BASIC ##
@attr.s(frozen=True, slots=True)
class BasicQOSPayload(MethodPayload):
    """
    Payload for the ``qos`` method.
    """

    klass = ClassID.BASIC
    method = 10
    is_client_side = False

    #: The prefetch size, in octets.
    prefetch_size: int = attr.ib(metadata=aq_type("long"))

    #: The prefetch count, in messages.
    prefetch_count: int = attr.ib(metadata=aq_type("short"))

    #: If this is a global QOS, or a per-channel one.
    global_: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicQOSOkPayload(MethodPayload):
    """
    Payload for the ``qos-ok`` method.
    """

    klass = ClassID.BASIC
    method = 11
    is_client_side = True

    # empty body


@attr.s(frozen=True, slots=True)
class BasicConsumePayload(MethodPayload):
    """
    Payload for the ``consume`` method.
    """

    klass = ClassID.BASIC
    method = 20
    is_client_side = False

    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the queue to consume from.
    queue_name: str = attr.ib()

    #: The consumer tag to use. Will be automatically generated by the server if left blank.
    consumer_tag: str = attr.ib()

    # todo doc
    no_wait: bool = attr.ib()
    exclusive: bool = attr.ib()
    no_ack: bool = attr.ib()
    no_local: bool = attr.ib()

    #: Extra, implementation specific arguments.
    arguments: Dict[str, Any] = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicConsumeOkPayload(MethodPayload):
    """
    Payload for the ``consume-ok`` method.
    """

    klass = ClassID.BASIC
    method = 21
    is_client_side = True

    #: The consumer tag for the consumer.
    consumer_tag: str = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicCancelPayload(MethodPayload):
    """
    Payload for the ``cancel`` method.
    """

    klass = ClassID.BASIC
    method = 30
    is_client_side = False

    #: The consumer tag to be cancelled.
    consumer_tag: str = attr.ib()

    #: If set, the server should not respond.
    no_wait: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicCancelOkPayload(MethodPayload):
    """
    Payload for the ``cancel-ok`` method.
    """

    klass = ClassID.BASIC
    method = 31
    is_client_side = True

    #: The consumer tag being cancelled.
    consumer_tag: str = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicPublishPayload(MethodPayload):
    """
    Payload for the ``publish`` method.
    """

    klass = ClassID.BASIC
    method = 40
    is_client_side = False

    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the exchange to publish the data to.
    name: str = attr.ib()

    #: The routing key to use.
    routing_key: str = attr.ib()

    #: If True, the server will return a Return message if the message cannot be sent to a consumer
    #: immediately.
    immediate: bool = attr.ib()

    #: If True, the server will return a Return message if the message is unrouteable.
    mandatory: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicReturnPayload(MethodPayload):
    """
    Payload for the ``return`` method.
    """

    klass = ClassID.BASIC
    method = 50
    is_client_side = True

    #: The code for the error that caused this message to be returned.
    reply_code: ReplyCode = attr.ib(converter=ReplyCode, metadata=aq_type("short"))

    #: The text for the error that caused this message to be returned.
    reply_text: str = attr.ib()

    #: The name of the exchange this was returned from.
    exchange: str = attr.ib()

    #: The message routing key.
    routing_key: str = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicDeliverPayload(MethodPayload):
    """
    Payload for the ``deliver`` method.
    """

    klass = ClassID.BASIC
    method = 60
    is_client_side = True

    #: The identifier for the consumer.
    consumer_tag: str = attr.ib()

    #: The server-assigned delivery tag.
    delivery_tag: int = attr.ib(metadata=aq_type("longlong"))

    #: Indicates that the message has been previously delivered.
    redelivered: bool = attr.ib()

    #: The name of the exchange the message was originally published to.
    exchange_name: str = attr.ib()

    #: The routing key for the message.
    routing_key: str = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicGetPayload(MethodPayload):
    """
    Payload for the ``get`` method.
    """

    klass = ClassID.BASIC
    method = 70
    is_client_side = False

    reserved_1: int = attr.ib(metadata=aq_type("short"))

    #: The name of the queue to get from.
    queue_name: str = attr.ib()

    #: If True, then messages will not be expected to be acknowledged.
    no_ack: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicGetOkPayload(MethodPayload):
    """
    Payload for the ``get-ok`` method.
    """

    klass = ClassID.BASIC
    method = 71
    is_client_side = True

    #: The server-assigned delivery tag.
    delivery_tag: int = attr.ib(metadata=aq_type("longlong"))

    #: Indicates that the message has been previously delivered.
    redelivered: bool = attr.ib()

    #: The name of the exchange the message was originally published to.
    exchange_name: str = attr.ib()

    #: The routing key for the message.
    routing_key: str = attr.ib()

    #: The message count remaining for the queue.
    message_count: int = attr.ib(metadata=aq_type("long"))


@attr.s(frozen=True, slots=True)
class BasicGetEmptyPayload(MethodPayload):
    """
    Payload for the ``get-empty`` method.
    """

    klass = ClassID.BASIC
    method = 72
    is_client_side = True

    reserved_1: str = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicAckPayload(MethodPayload):
    """
    Payload for the ``ack`` method.
    """

    klass = ClassID.BASIC
    method = 80
    is_client_side = True

    #: The server-assigned delivery tag.
    delivery_tag: int = attr.ib(metadata=aq_type("longlong"))

    #: If True, the delivery tag is set to "up to and including".
    multiple: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicRejectPayload(MethodPayload):
    """
    Payload for the ``reject`` method.
    """

    klass = ClassID.BASIC
    method = 90
    is_client_side = False

    #: The server-assigned delivery tag.
    delivery_tag: int = attr.ib(metadata=aq_type("longlong"))

    #: If True, the message should be re-queued.
    requeue: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class BasicNackPayload(MethodPayload):
    """
    Payload for the ``nack`` method.
    """

    klass = ClassID.BASIC
    method = 120
    is_client_side = True

    #: The server-assigned delivery tag.
    delivery_tag: int = attr.ib(metadata=aq_type("longlong"))

    #: If True, the message should be re-queued.
    requeue: bool = attr.ib()

    #: If True, the delivery tag is set to "up to and including".
    multiple: bool = attr.ib()


## CONFIRM (RabbitMQ extension) ##
@attr.s(frozen=True, slots=True)
class ConfirmSelectPayload(MethodPayload):
    """
    Payload for the ``select`` method.
    """

    klass = ClassID.CONFIRM
    method = 10
    is_client_side = False

    #: If set, the server should not respond.
    no_wait: bool = attr.ib()


@attr.s(frozen=True, slots=True)
class ConfirmSelectOkPayload(MethodPayload):
    """
    Payload for the ``select-ok`` method.
    """

    klass = ClassID.CONFIRM
    method = 11
    is_client_side = True

    # empty body


PAYLOAD_TYPES = {
    ClassID.CONNECTION: {
        ConnectionStartPayload.method: ConnectionStartPayload,
        ConnectionStartOkPayload.method: ConnectionStartOkPayload,
        ConnectionSecurePayload.method: ConnectionSecurePayload,
        ConnectionSecureOkPayload.method: ConnectionSecureOkPayload,
        ConnectionTunePayload.method: ConnectionTunePayload,
        ConnectionTuneOkPayload.method: ConnectionTuneOkPayload,
        ConnectionOpenPayload.method: ConnectionOpenPayload,
        ConnectionOpenOkPayload.method: ConnectionOpenOkPayload,
        ConnectionClosePayload.method: ConnectionClosePayload,
        ConnectionCloseOkPayload.method: ConnectionCloseOkPayload,
    },
    ClassID.CHANNEL: {
        ChannelOpenPayload.method: ChannelOpenPayload,
        ChannelOpenOkPayload.method: ChannelOpenOkPayload,
        ChannelFlowPayload.method: ChannelFlowPayload,
        ChannelFlowOkPayload.method: ChannelFlowOkPayload,
        ChannelClosePayload.method: ChannelClosePayload,
        ChannelCloseOkPayload.method: ChannelCloseOkPayload,
    },
    ClassID.EXCHANGE: {
        ExchangeDeclarePayload.method: ExchangeDeclarePayload,
        ExchangeDeclareOkPayload.method: ExchangeDeclareOkPayload,
        ExchangeDeletePayload.method: ExchangeDeletePayload,
        ExchangeDeleteOkPayload.method: ExchangeDeleteOkPayload,
        ExchangeBindPayload.method: ExchangeBindPayload,
        ExchangeBindOkPayload.method: ExchangeBindOkPayload,
        ExchangeUnBindPayload.method: ExchangeUnBindPayload,
        ExchangeUnBindOkPayload.method: ExchangeUnBindPayload,
    },
    ClassID.QUEUE: {
        QueueDeclarePayload.method: QueueDeclarePayload,
        QueueDeclareOkPayload.method: QueueDeclareOkPayload,
        QueueBindPayload.method: QueueBindPayload,
        QueueBindOkPayload.method: QueueBindOkPayload,
        QueueDeletePayload.method: QueueDeletePayload,
        QueueDeleteOkPayload.method: QueueDeleteOkPayload,
        QueuePurgePayload.method: QueuePurgePayload,
        QueuePurgeOkPayload.method: QueuePurgeOkPayload,
        QueueUnbindPayload.method: QueueUnbindPayload,
        QueueUnbindOkPayload.method: QueueUnbindOkPayload,
    },
    ClassID.BASIC: {
        BasicPublishPayload.method: BasicPublishPayload,
        BasicDeliverPayload.method: BasicDeliverPayload,
        BasicConsumePayload.method: BasicConsumePayload,
        BasicConsumeOkPayload.method: BasicConsumeOkPayload,
        BasicCancelPayload.method: BasicCancelPayload,
        BasicCancelOkPayload.method: BasicCancelOkPayload,
        BasicQOSPayload.method: BasicQOSPayload,
        BasicQOSOkPayload.method: BasicQOSOkPayload,
        BasicAckPayload.method: BasicAckPayload,
        BasicGetPayload.method: BasicGetPayload,
        BasicGetOkPayload.method: BasicGetOkPayload,
        BasicGetEmptyPayload.method: BasicGetEmptyPayload,
        BasicReturnPayload.method: BasicReturnPayload,
    },
    ClassID.CONFIRM: {
        ConfirmSelectPayload.method: ConfirmSelectPayload,
        ConfirmSelectOkPayload.method: ConfirmSelectOkPayload,
    },
}


def deserialise_payload(body: bytes) -> MethodPayload:
    """
    Deserialises a method payload.

    :param body: The body of the payload to decode.
    :return: A :class:`.MethodPayload` matching the returned payload.
    """

    klass, method, rest = (
        ClassID(int.from_bytes(body[0:2], byteorder="big")),
        int.from_bytes(body[2:4], byteorder="big"),
        body[4:],
    )

    try:
        payload_klass = PAYLOAD_TYPES[klass][method]
    except KeyError:
        raise KeyError(f"Unknown method: {klass.name}/{method}") from None

    attr.resolve_types(payload_klass)
    fields = attr.fields(payload_klass)
    init_params = {}
    buffer = DecodingBuffer(rest)

    for field in fields:
        init_params[field.name] = decode_attrs_attribute(buffer, field)

    return payload_klass(**init_params)  # type: ignore


def serialise_payload(payload: MethodPayload) -> bytes:
    """
    Serialises a payload into a bytearray.
    """

    typ = type(payload)
    attr.resolve_types(typ)

    header = typ.klass.to_bytes(2, byteorder="big") + typ.method.to_bytes(2, byteorder="big")

    buf = EncodingBuffer()
    fields = attr.fields(typ)
    for field in fields:
        encode_attrs_attribute(buf, field, getattr(payload, field.name))

    buf.force_write_bits()
    return header + buf.get_data()


def method_payload_name(payload: MethodPayload):
    """
    Gets the name of a method payload.
    """

    return f"{payload.klass.name}/{payload.method}/{type(payload).__name__}"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['serena', 'serena.payloads', 'serena.utils']

package_data = \
{'': ['*']}

install_requires = \
['anyio>=3.5.0,<4.0.0', 'attrs>=21.4.0,<22.0.0']

extras_require = \
{':python_version < "3.11"': ['backports.strenum'],
 'docs': ['Sphinx>=4.3.2,<5.0.0',
          'sphinx-rtd-theme>=1.0.0,<2.0.0',
          'sphinx-inline-tabs>=2022.1.2-beta.11,<2023.0.0',
          'sphinxcontrib-trio>=1.1.2,<2.0.0',
          'sphinx-autodoc-typehints>=1.15.2,<2.0.0']}

setup_kwargs = {
    'name': 'serena',
    'version': '0.7.0',
    'description': 'An AMQP 0-9-1 client using AnyIO.',
    'long_description': 'Serena\n======\n\n*Serena* is a pure-Python, structually concurrent AMQP 0-9-1 client.\n\nSee https://amqp.py.veriny.tf/ for more information.\n',
    'author': 'Lura Skye',
    'author_email': 'l@veriny.tf',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
